from rich.console import Console
from rich.style import Style

console = Console()

warning_style = Style(color="yellow", bold=True)
success_style = Style(color="green", bold=True)
danger_style = Style(color="red", bold=True)